package ru.dbotthepony.kstarbound.server

import com.google.gson.JsonPrimitive
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.SupervisorJob
import kotlinx.coroutines.async
import kotlinx.coroutines.cancel
import kotlinx.coroutines.future.asCompletableFuture
import kotlinx.coroutines.future.await
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.vector.Vector3i
import ru.dbotthepony.kstarbound.Globals
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.WorldID
import ru.dbotthepony.kstarbound.defs.world.AsteroidsWorldParameters
import ru.dbotthepony.kstarbound.defs.world.FloatingDungeonWorldParameters
import ru.dbotthepony.kstarbound.defs.world.TerrestrialWorldParameters
import ru.dbotthepony.kstarbound.defs.world.WorldTemplate
import ru.dbotthepony.kstarbound.network.packets.clientbound.UniverseTimeUpdatePacket
import ru.dbotthepony.kstarbound.server.world.ServerUniverse
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.server.world.ServerSystemWorld
import ru.dbotthepony.kstarbound.server.world.WorldStorage
import ru.dbotthepony.kstarbound.util.Clock
import ru.dbotthepony.kstarbound.util.ExceptionLogger
import ru.dbotthepony.kstarbound.util.ExecutionSpinner
import ru.dbotthepony.kstarbound.util.MailboxExecutorService
import ru.dbotthepony.kstarbound.util.random.random
import ru.dbotthepony.kstarbound.world.UniversePos
import java.io.Closeable
import java.io.File
import java.util.UUID
import java.util.concurrent.CompletableFuture
import java.util.concurrent.TimeUnit
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Supplier

sealed class StarboundServer(val root: File) : Closeable {
	init {
		if (!root.exists()) {
			check(root.mkdirs()) { "Unable to create ${root.absolutePath}" }
		} else if (!root.isDirectory) {
			throw IllegalArgumentException("${root.absolutePath} is not a directory")
		}
	}

	private val worlds = HashMap<WorldID, CompletableFuture<ServerWorld>>()
	val mailbox = MailboxExecutorService().also { it.exceptionHandler = ExceptionLogger(LOGGER) }
	val spinner = ExecutionSpinner(mailbox::executeQueuedTasks, ::tick, Starbound.TIMESTEP_NANOS)
	val thread = Thread(spinner, "Server Thread")
	val universe = ServerUniverse()
	val chat = ChatHandler(this)
	val scope = CoroutineScope(Starbound.COROUTINE_EXECUTOR + SupervisorJob())

	private val systemWorlds = HashMap<Vector3i, CompletableFuture<ServerSystemWorld>>()

	private suspend fun loadSystemWorld0(location: Vector3i): ServerSystemWorld {
		return ServerSystemWorld.create(this, location)
	}

	fun loadSystemWorld(location: Vector3i): CompletableFuture<ServerSystemWorld> {
		return CompletableFuture.supplyAsync(Supplier {
			systemWorlds.computeIfAbsent(location) {
				scope.async { loadSystemWorld0(location) }.asCompletableFuture()
			}
		}, mailbox).thenCompose { it }
	}

	private suspend fun loadCelestialWorld(location: WorldID.Celestial): ServerWorld {
		LOGGER.info("Creating celestial world $location")

		val template = WorldTemplate.create(location.pos, universe)
		val world = ServerWorld.create(this, template, WorldStorage.Nothing, location)

		try {
			world.thread.start()
			world.prepare().await()
		} catch (err: Throwable) {
			LOGGER.fatal("Exception while creating celestial world at $location!", err)
			world.close()
			throw err
		}

		return world
	}

	private suspend fun loadInstanceWorld(location: WorldID.Instance): ServerWorld {
		val config = Globals.instanceWorlds[location.name] ?: throw NoSuchElementException("No such instance world ${location.name}")
		val random = random(config.seed ?: System.nanoTime())

		val visitable = when (config.type.lowercase()) {
			"terrestrial" -> TerrestrialWorldParameters.generate(config.planetType!!, config.planetSize!!, random)
			"asteroids" -> AsteroidsWorldParameters.generate(random)
			"floatingdungeon" -> FloatingDungeonWorldParameters.generate(config.dungeonWorld!!)
			else -> throw RuntimeException()
		}

		if (location.threatLevel != null) {
			visitable.threatLevel = location.threatLevel
		}

		if (config.beamUpRule != null) {
			visitable.beamUpRule = config.beamUpRule
		}

		visitable.disableDeathDrops = config.disableDeathDrops

		val template = WorldTemplate(visitable, config.skyParameters, random)
		val world = ServerWorld.create(this, template, WorldStorage.NULL, location)

		world.setProperty("ephemeral", JsonPrimitive(!config.persistent))

		world.thread.start()
		return world
	}

	private suspend fun loadWorld0(location: WorldID): ServerWorld {
		return when (location) {
			is WorldID.ShipWorld -> throw IllegalArgumentException("Can't create ship worlds out of thin air")
			is WorldID.Instance -> loadInstanceWorld(location)
			is WorldID.Celestial -> loadCelestialWorld(location)
			is WorldID.Limbo -> throw IllegalArgumentException("Limbo was supplied as world ID")
		}
	}

	fun loadWorld(location: WorldID): CompletableFuture<ServerWorld> {
		return CompletableFuture.supplyAsync(Supplier {
			var world = worlds[location]

			if (world != null && world.isCompletedExceptionally) {
				worlds.remove(location)
				world = null
			}

			if (world != null) {
				world
			} else {
				val future = scope.async { loadWorld0(location) }.asCompletableFuture()
				worlds[location] = future
				future
			}
		}, mailbox).thenCompose { it }
	}

	fun loadShipWorld(connection: ServerConnection, storage: WorldStorage): CompletableFuture<ServerWorld> {
		return CompletableFuture.supplyAsync(Supplier {
			val id = WorldID.ShipWorld(connection.uuid ?: throw NullPointerException("Connection UUID is null"))
			val existing = worlds[id]

			if (existing != null)
				throw IllegalStateException("Already has $id!")

			val world = ServerWorld.load(this, storage, id)
			worlds[id] = world
			world
		}, mailbox).thenCompose { it }
	}

	fun notifyWorldUnloaded(worldID: WorldID) {
		mailbox.execute {
			worlds.remove(worldID)
		}
	}

	fun loadSystemWorld(location: UniversePos): CompletableFuture<ServerSystemWorld> {
		return loadSystemWorld(location.location)
	}

	val settings = ServerSettings()
	val channels = ServerChannels(this)
	val lock = ReentrantLock()
	var isClosed = false
		private set

	var serverUUID: UUID = UUID.randomUUID()
		protected set

	val universeClock = Clock()

	init {
		mailbox.scheduleAtFixedRate(Runnable {
			channels.broadcast(UniverseTimeUpdatePacket(universeClock.seconds))
		}, Globals.universeServer.clockUpdatePacketInterval, Globals.universeServer.clockUpdatePacketInterval, TimeUnit.MILLISECONDS)

		thread.uncaughtExceptionHandler = Thread.UncaughtExceptionHandler { t, e ->
			LOGGER.fatal("Unexpected exception in server execution loop, shutting down", e)
			actuallyClose()
		}

		// thread.isDaemon = this is IntegratedStarboundServer
		thread.start()
	}

	private val occupiedNicknames = ObjectArraySet<String>()

	fun reserveNickname(name: String, alternative: String): String {
		synchronized(occupiedNicknames) {
			var name = name

			if (name.lowercase() == "server" || name.isBlank()) {
				name = alternative
			}

			while (name in occupiedNicknames) {
				name += "_"
			}

			occupiedNicknames.add(name)
			return name
		}
	}

	fun freeNickname(name: String): Boolean {
		return synchronized(occupiedNicknames) {
			occupiedNicknames.remove(name)
		}
	}

	fun clientByUUID(uuid: UUID): ServerConnection? {
		return channels.connections.firstOrNull { it.uuid == uuid }
	}

	protected abstract fun close0()
	protected abstract fun tick0()

	private fun tick(): Boolean {
		if (isClosed) return false

		channels.connections.forEach {
			try {
				it.tick()
			} catch (err: Throwable) {
				LOGGER.error("Exception while ticking client connection", err)
				it.disconnect("Exception while ticking client connection: $err")
			}
		}

		// TODO: schedule to thread pool?
		// right now, system worlds are rather lightweight, and having separate threads for them is overkill
		if (systemWorlds.isNotEmpty()) {
			runBlocking {
				systemWorlds.values.removeIf {
					if (it.isCompletedExceptionally) {
						return@removeIf true
					}

					if (!it.isDone) {
						return@removeIf false
					}

					launch { it.get().tick() }

					if (it.get().shouldClose()) {
						LOGGER.info("Stopping idling ${it.get()}")
						return@removeIf true
					}

					return@removeIf false
				}
			}
		}

		tick0()
		return !isClosed
	}

	private fun actuallyClose() {
		if (isClosed) return
		isClosed = true

		scope.cancel("Server shutting down")
		channels.close()

		worlds.values.forEach {
			if (it.isDone && !it.isCompletedExceptionally)
				it.get().close()

			it.cancel(true)
		}

		universe.close()
		close0()
	}

	final override fun close() {
		if (Thread.currentThread() == thread) {
			actuallyClose()
		} else {
			mailbox.execute { actuallyClose() }
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
