package ru.dbotthepony.kstarbound.server.world

import com.google.gson.JsonElement
import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.ints.IntArraySet
import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.SupervisorJob
import kotlinx.coroutines.asCoroutineDispatcher
import kotlinx.coroutines.cancel
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.future.asCompletableFuture
import kotlinx.coroutines.future.await
import kotlinx.coroutines.launch
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.util.AABB
import ru.dbotthepony.kommons.util.AABBi
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Globals
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.WarpAction
import ru.dbotthepony.kstarbound.defs.WarpAlias
import ru.dbotthepony.kstarbound.defs.WorldID
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.defs.tile.TileDamageResult
import ru.dbotthepony.kstarbound.defs.tile.TileDamageType
import ru.dbotthepony.kstarbound.defs.tile.isEmptyLiquid
import ru.dbotthepony.kstarbound.defs.world.WorldStructure
import ru.dbotthepony.kstarbound.defs.world.WorldTemplate
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.jsonArrayOf
import ru.dbotthepony.kstarbound.network.IPacket
import ru.dbotthepony.kstarbound.network.packets.StepUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.PlayerWarpResultPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.UpdateWorldPropertiesPacket
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.server.ServerConnection
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.ExecutionSpinner
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.IChunkListener
import ru.dbotthepony.kstarbound.world.TileHealth
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.WorldGeometry
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.tile.TileEntity
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import java.util.concurrent.CompletableFuture
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.RejectedExecutionException
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.locks.LockSupport
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Predicate
import java.util.function.Supplier
import kotlin.concurrent.withLock
import kotlin.properties.Delegates

class ServerWorld private constructor(
	val server: StarboundServer,
	template: WorldTemplate,
	val storage: WorldStorage,
	val worldID: WorldID,
) : World<ServerWorld, ServerChunk>(template) {
	init {
		if (server.isClosed)
			throw RuntimeException()
	}

	val clients = CopyOnWriteArrayList<ServerWorldTracker>()
	val shouldStopOnIdle = worldID !is WorldID.ShipWorld
	val scope = CoroutineScope(mailbox.asCoroutineDispatcher() + SupervisorJob())

	private fun doAcceptClient(client: ServerConnection, action: WarpAction?) {
		try {
			isBusy++

			if (clients.any { it.client == client })
				throw IllegalStateException("$client is already in $this")

			if (!client.isConnected)
				throw IllegalStateException("$client disconnected while joining $this")

			val start = if (action is WarpAction.Player)
				clients.firstOrNull { it.client.uuid == action.uuid }?.client?.playerEntity?.position
			else if (action is WarpAction.World)
				action.target.resolve(this)
			else
				playerSpawnPosition

			if (start == null) {
				client.send(PlayerWarpResultPacket(false, action!!, true))
				throw IllegalStateException("Not a valid spawn target: $action")
			}

			if (action != null)
				client.send(PlayerWarpResultPacket(true, action, false))

			client.tracker?.remove("Transiting to new world")
			clients.add(ServerWorldTracker(this, client, start))
		} finally {
			isBusy--
		}
	}

	fun acceptClient(player: ServerConnection, action: WarpAction? = null): CompletableFuture<Unit> {
		check(!isClosed.get()) { "$this is invalid" }
		unpause()

		try {
			val future = CompletableFuture.supplyAsync(Supplier { doAcceptClient(player, action) }, mailbox)

			future.exceptionally {
				LOGGER.error("Error while accepting new player into world", it)
			}

			return future
		} catch (err: RejectedExecutionException) {
			return CompletableFuture.failedFuture(err)
		}
	}

	val spinner = ExecutionSpinner(mailbox::executeQueuedTasks, ::spin, Starbound.TIMESTEP_NANOS)
	private val str = "Server World ${worldID.toString()}"
	val thread = Thread(spinner, str)

	init {
		mailbox.thread = thread
	}

	private val isClosed = AtomicBoolean()

	fun isClosed(): Boolean {
		return isClosed.get()
	}

	init {
		thread.isDaemon = true
	}

	fun pause() {
		if (!isClosed.get()) spinner.pause()
	}

	fun unpause() {
		if (!isClosed.get()) spinner.unpause()
	}

	override fun toString(): String {
		if (isClosed.get())
			return "NULL $str"
		else
			return str
	}

	override fun close() {
		if (!isClosed.get())
			LOGGER.info("Shutting down $this")

		if (isClosed.compareAndSet(false, true)) {
			server.notifyWorldUnloaded(worldID)
			super.close()
			spinner.unpause()

			ticketListLock.withLock {
				ticketLists.forEach { it.scope.cancel() }
			}

			clients.forEach {
				it.remove()
				it.client.enqueueWarp(WarpAlias.Return)
			}

			LockSupport.unpark(thread)
		}
	}

	private var idleTicks = 0
	private var isBusy = 0

	private fun spin(): Boolean {
		if (isClosed.get()) return false

		try {
			if (clients.isEmpty() && isBusy <= 0) {
				idleTicks++
			} else {
				idleTicks = 0
			}

			tick()

			if (idleTicks >= 600) {
				if (shouldStopOnIdle) {
					close()
					return false
				} else {
					pause()
				}
			}

			return true
		} catch (err: Throwable) {
			LOGGER.fatal("Exception in world tick loop", err)
			close()
			return false
		}
	}

	override val isRemote: Boolean
		get() = false

	override fun isSameThread(): Boolean {
		return Thread.currentThread() === thread
	}

	fun damageTiles(positions: Collection<IStruct2i>, isBackground: Boolean, sourcePosition: Vector2d, damage: TileDamage, source: AbstractEntity? = null): TileDamageResult {
		if (damage.amount <= 0.0)
			return TileDamageResult.NONE

		val actualPositions = positions.stream()
			.map { geometry.wrap(it) }
			.distinct()
			.map { it to chunkMap[geometry.chunkFromCell(it)] }
			.toList()

		var topMost = TileDamageResult.NONE

		val damagedEntities = ObjectArraySet<TileEntity>()

		for ((pos, chunk) in actualPositions) {
			var damage = damage
			var tileEntityResult = TileDamageResult.NONE

			if (chunk?.getCell(pos - chunk.pos.tile)?.dungeonId in protectedDungeonIDs)
				damage = damage.copy(type = TileDamageType.PROTECTED)

			if (!isBackground) {
				for (entity in entitiesAtTile(pos, distinct = false)) {
					if (!damagedEntities.add(entity)) continue

					val occupySpaces = entity.occupySpaces.stream()
						.map { geometry.wrap(it + entity.tilePosition) }
						.filter { it in positions }
						.toList()

					val broken = entity.damage(occupySpaces, sourcePosition, damage)

					if (source != null && broken) {
						source.receiveMessage("tileEntityBroken", jsonArrayOf(pos, entity.type.jsonName, (entity as? WorldObject)?.config?.key))
					}

					if (damage.type == TileDamageType.PROTECTED)
						tileEntityResult = TileDamageResult.PROTECTED
					else
						tileEntityResult = TileDamageResult.NORMAL
				}
			}

			// Penetrating damage should carry through to the blocks behind this
			// entity.
			if (tileEntityResult == TileDamageResult.NONE || damage.type.isPenetrating) {
				chunk ?: continue
				val (result, health, tile) = chunk.damageTile(pos - chunk.pos.tile, isBackground, sourcePosition, damage, source)
				topMost = topMost.coerceAtLeast(result)

				if (source != null && health?.isDead == true) {
					source.receiveMessage("tileBroken", jsonArrayOf(
						pos, if (isBackground) "background" else "foreground",
						tile!!.tile(isBackground).material.id ?: 0, // TODO: string identifiers support
						tile.dungeonId,
						health.isHarvested
					))
				}
			}

			topMost = topMost.coerceAtLeast(tileEntityResult)
		}

		return topMost
	}

	override fun tick() {
		super.tick()
		val packet = StepUpdatePacket(ticks)

		clients.forEach {
			if (!isClosed.get()) {
				it.send(packet)

				try {
					it.tick()
				} catch (err: Throwable) {
					LOGGER.error("Exception while ticking player ${it.client}", err)
					//it.disconnect("Exception while ticking player: $err")
				}
			}
		}

		ticketListLock.withLock {
			ticketLists.removeIf { it.tick() }
		}
	}

	override fun broadcast(packet: IPacket) {
		clients.forEach {
			it.send(packet)
		}
	}

	private suspend fun prepare0() {
		try {
			isBusy++

			if (playerSpawnPosition == Vector2d.ZERO) {
				playerSpawnPosition = findPlayerStart()
			}
		} finally {
			isBusy--
		}
	}

	// done as completable future because we must do
	// everything inside our own thread, not anywhere else
	// This way, external callers can properly wait for preparations to complete
	fun prepare(): CompletableFuture<*> {
		return CompletableFuture.supplyAsync(Supplier {
			scope.launch { prepare0() }.asCompletableFuture()
		}, mailbox).thenCompose { it }
	}

	private suspend fun findPlayerStart(hint: Vector2d? = null): Vector2d {
		val tickets = ArrayList<ITicket>()

		try {
			LOGGER.info("Trying to find player spawn position...")
			var pos = hint ?: CompletableFuture.supplyAsync(Supplier { template.findSensiblePlayerStart() }, Starbound.EXECUTOR).await() ?: Vector2d(0.0, template.surfaceLevel().toDouble())
			var previous = pos
			LOGGER.info("Trying to find player spawn position near $pos...")

			for (t in 0 until Globals.worldServer.playerStartRegionMaximumTries) {
				var foundGround = false

				// First go downward until we collide with terrain
				for (i in 0 until Globals.worldServer.playerStartRegionMaximumVerticalSearch) {
					val spawnRect = AABB(
						Vector2d(pos.x - Globals.worldServer.playerStartRegionSize.x / 2, pos.y),
						Vector2d(pos.x + Globals.worldServer.playerStartRegionSize.x / 2, pos.y + Globals.worldServer.playerStartRegionSize.y),
					)

					val region = permanentChunkTicket(spawnRect)
					tickets.addAll(region)
					region.forEach { it.chunk.await() }

					foundGround = matchCells(spawnRect) {
						it.foreground.material.value.collisionKind != CollisionType.NONE
					}

					if (foundGround) {
						break
					} else {
						pos += Vector2d.NEGATIVE_Y
					}
				}

				if (foundGround) {
					// Then go up until our spawn region is no longer in the terrain, but bail
					// out and try again if we can't signal the region or we are stuck in a
					// dungeon.

					for (i in 0 until Globals.worldServer.playerStartRegionMaximumVerticalSearch) {
						if (!chunkMap.getCell(pos.x.toInt(), pos.y.toInt()).liquid.state.isEmptyLiquid) {
							break
						}

						val spawnRect = AABB(
							Vector2d(pos.x - Globals.worldServer.playerStartRegionSize.x / 2, pos.y),
							Vector2d(pos.x + Globals.worldServer.playerStartRegionSize.x / 2, pos.y + Globals.worldServer.playerStartRegionSize.y),
						)

						val region = permanentChunkTicket(spawnRect)
						tickets.addAll(region)
						region.forEach { it.chunk.await() }

						if (!matchCells(spawnRect) { it.foreground.material.value.collisionKind != CollisionType.NONE } && spawnRect.maxs.y < geometry.size.y) {
							LOGGER.info("Found appropriate spawn position at $pos")
							return pos
						}

						pos += Vector2d.POSITIVE_Y
					}
				}

				pos = CompletableFuture.supplyAsync(Supplier { template.findSensiblePlayerStart() }, Starbound.EXECUTOR).await() ?: Vector2d(0.0, template.surfaceLevel().toDouble())

				if (previous != pos) {
					LOGGER.info("Still trying to find player spawn position near $pos...")
					previous = pos
				} else {
					break
				}
			}

			LOGGER.warn("Unable to find proper player start location, will use $pos")
			return pos
		} finally {
			tickets.forEach { it.cancel() }
		}
	}

	override fun setProperty0(key: String, value: JsonElement) {
		super.setProperty0(key, value)
		broadcast(UpdateWorldPropertiesPacket(JsonObject().apply { add(key, value) }))
	}

	override fun chunkFactory(pos: ChunkPos): ServerChunk {
		return ServerChunk(this, pos)
	}

	private val ticketMap = Long2ObjectOpenHashMap<TicketList>()
	private val ticketLists = ArrayList<TicketList>()
	private val ticketListLock = ReentrantLock()

	private fun getTicketList(pos: ChunkPos): TicketList {
		return ticketMap.computeIfAbsent(geometry.wrapToLong(pos), Long2ObjectFunction { TicketList(it) })
	}

	fun permanentChunkTicket(pos: ChunkPos, target: ServerChunk.State = ServerChunk.State.FULL): ITicket {
		ticketListLock.withLock {
			return getTicketList(pos).Ticket(target)
		}
	}

	fun permanentChunkTicket(region: AABBi, target: ServerChunk.State = ServerChunk.State.FULL): List<ITicket> {
		ticketListLock.withLock {
			return geometry.region2Chunks(region).map { getTicketList(it).Ticket(target) }
		}
	}

	fun permanentChunkTicket(region: AABB, target: ServerChunk.State = ServerChunk.State.FULL): List<ITicket> {
		ticketListLock.withLock {
			return geometry.region2Chunks(region).map { getTicketList(it).Ticket(target) }
		}
	}

	fun temporaryChunkTicket(pos: ChunkPos, time: Int, target: ServerChunk.State = ServerChunk.State.FULL): ITimedTicket {
		require(time > 0) { "Invalid ticket time: $time" }

		ticketListLock.withLock {
			return getTicketList(pos).TimedTicket(time, target)
		}
	}

	fun temporaryChunkTicket(region: AABBi, time: Int, target: ServerChunk.State = ServerChunk.State.FULL): List<ITimedTicket> {
		require(time > 0) { "Invalid ticket time: $time" }

		ticketListLock.withLock {
			return geometry.region2Chunks(region).map { getTicketList(it).TimedTicket(time, target) }
		}
	}

	fun temporaryChunkTicket(region: AABB, time: Int, target: ServerChunk.State = ServerChunk.State.FULL): List<ITimedTicket> {
		require(time > 0) { "Invalid ticket time: $time" }

		ticketListLock.withLock {
			return geometry.region2Chunks(region).map { getTicketList(it).TimedTicket(time, target) }
		}
	}

	override fun onChunkCreated(chunk: ServerChunk) {
		ticketListLock.withLock {
			ticketMap[chunk.pos.toLong()]?.let { chunk.addListener(it) }
		}
	}

	override fun onChunkRemoved(chunk: ServerChunk) {
		ticketListLock.withLock {
			ticketMap[chunk.pos.toLong()]?.let { chunk.removeListener(it) }
		}
	}

	interface ITicket {
		fun cancel()
		val isCanceled: Boolean
		val pos: ChunkPos
		val id: Int
		val chunk: CompletableFuture<ServerChunk>
		var listener: IChunkListener?
	}

	interface ITimedTicket : ITicket, Comparable<ITimedTicket> {
		val timeRemaining: Int
		fun prolong(ticks: Int)

		override fun compareTo(other: ITimedTicket): Int {
			val cmp = timeRemaining.compareTo(other.timeRemaining)
			if (cmp != 0) return cmp
			return id.compareTo(other.id)
		}
	}

	private inner class TicketList(val pos: ChunkPos) : IChunkListener {
		constructor(pos: Long) : this(ChunkPos(pos))

		private var calledLoadChunk = true
		private val permanent = ArrayList<Ticket>()
		private val temporary = ObjectAVLTreeSet<TimedTicket>()
		private var ticks = 0
		private var nextTicketID = 0
		private var isBusy = false
		private var chunk by Delegates.notNull<ServerChunk>()
		private val targetState = Channel<ServerChunk.State>(Int.MAX_VALUE)
		val scope = CoroutineScope(mailbox.asCoroutineDispatcher())
		private var idleTicks = 0
		private var isRemoved = false

		private suspend fun chunkGeneratorLoop() {
			while (true) {
				if (chunk.state == ServerChunk.State.FULL) {
					break
				}

				val targetState = targetState.receive()

				while (chunk.state < targetState) {
					isBusy = true

					val nextState = ServerChunk.State.entries[chunk.state.ordinal + 1]

					try {
						when (nextState) {
							ServerChunk.State.TILES -> {
								// tiles can be generated concurrently without any consequences
								CompletableFuture.runAsync(Runnable { chunk.prepareCells() }, Starbound.EXECUTOR).await()
							}

							ServerChunk.State.MICRO_DUNGEONS -> {
								//LOGGER.error("NYI: Generating microdungeons for $chunk")
							}

							ServerChunk.State.CAVE_LIQUID -> {
								//LOGGER.error("NYI: Generating cave liquids for $chunk")
							}

							ServerChunk.State.TILE_ENTITIES -> {
								//LOGGER.error("NYI: Generating tile entities for $chunk")
							}

							ServerChunk.State.ENTITIES -> {
								//LOGGER.error("NYI: Placing entities for $chunk")
							}

							else -> {}
						}

						chunk.bumpState(nextState)
						fulfilFutures()
					} catch (err: Throwable) {
						LOGGER.error("Exception while propagating $chunk to next generation state $nextState", err)
						break
					}
				}

				isBusy = false
			}

			isBusy = false
		}

		private suspend fun loadChunk0() {
			try {
				val cells = storage.loadCells(pos).await()

				// very good.
				if (cells.isPresent) {
					chunk.loadCells(cells.value)
					chunk.bumpState(ServerChunk.State.CAVE_LIQUID)
					fulfilFutures()

					storage.loadEntities(pos).await().ifPresent {
						for (obj in it) {
							obj.joinWorld(this@ServerWorld)
						}
					}

					chunk.bumpState(ServerChunk.State.FULL)
					fulfilFutures()
					isBusy = false
					return
				} else {
					// generate.
					chunkGeneratorLoop()
				}
			} catch (err: Throwable) {
				LOGGER.error("Exception while loading chunk $chunk", err)
			}
		}

		private fun loadChunk() {
			if (!calledLoadChunk)
				return

			calledLoadChunk = true

			if (geometry.x.inBoundsChunk(pos.x) && geometry.y.inBoundsChunk(pos.y)) {
				ticketListLock.withLock {
					ticketLists.add(this)
				}

				val existing = chunkMap[pos]

				if (existing == null) {
					// fresh chunk
					val chunk = chunkMap.compute(pos) ?: return ticketListLock.withLock {
						isRemoved = true
						ticketLists.remove(this)
						ticketMap.remove(pos.toLong())
					}

					this.chunk = chunk

					chunk.addListener(this)
					isBusy = true
					scope.launch { loadChunk0() }
					fulfilFutures()
				} else {
					chunk = existing
					existing.addListener(this)
					fulfilFutures()
				}
			}
		}

		private fun unload() {
			if (isRemoved)
				return

			isRemoved = true
			scope.cancel()
			targetState.close()

			val removed = ticketMap.remove(pos.toLong())
			check(removed == this) { "Expected to remove $this, but removed $removed" }

			if (chunk.state == ServerChunk.State.FULL) {
				val unloadable = entityIndex
					.query(
						chunk.aabb,
						filter = Predicate { it.isApplicableForUnloading && chunk.aabb.isInside(it.position) },
						distinct = true, withEdges = false)

				storage.saveCells(pos, chunk.copyCells())
				storage.saveEntities(pos, unloadable)

				unloadable.forEach {
					it.remove()
				}
			}

			chunkMap.remove(pos)
		}

		fun tick(): Boolean {
			ticks++

			while (temporary.isNotEmpty() && temporary.first().timeRemaining <= 0) {
				val ticket = temporary.first()
				ticket.isCanceled = true
				temporary.remove(ticket)
			}

			var shouldUnload = !isBusy && temporary.isEmpty() && permanent.isEmpty()

			if (shouldUnload) {
				idleTicks++
				// don't load-save-load-save too frequently
				shouldUnload = idleTicks > 600
			} else {
				idleTicks = 0
			}

			if (shouldUnload) {
				unload()
			}

			return shouldUnload
		}

		private fun fulfilFutures() {
			val permanent: List<Ticket>
			val temporary: List<Ticket>

			ticketListLock.withLock {
				permanent = ObjectArrayList(this.permanent)
				temporary = ObjectArrayList(this.permanent)
			}

			val state = chunk.state

			permanent.forEach { if (it.targetState <= state) it.chunk.complete(chunk) }
			temporary.forEach { if (it.targetState <= state) it.chunk.complete(chunk) }
		}

		override fun onCellChanges(x: Int, y: Int, cell: ImmutableCell) {
			val permanent: List<Ticket>
			val temporary: List<Ticket>

			ticketListLock.withLock {
				permanent = ObjectArrayList(this.permanent)
				temporary = ObjectArrayList(this.permanent)
			}

			val state = chunk.state

			permanent.forEach { if (it.targetState <= state) it.listener?.onCellChanges(x, y, cell) }
			temporary.forEach { if (it.targetState <= state) it.listener?.onCellChanges(x, y, cell) }
		}

		override fun onTileHealthUpdate(x: Int, y: Int, isBackground: Boolean, health: TileHealth) {
			val permanent: List<Ticket>
			val temporary: List<Ticket>

			ticketListLock.withLock {
				permanent = ObjectArrayList(this.permanent)
				temporary = ObjectArrayList(this.permanent)
			}

			val state = chunk.state

			permanent.forEach { if (it.targetState <= state) it.listener?.onTileHealthUpdate(x, y, isBackground, health) }
			temporary.forEach { if (it.targetState <= state) it.listener?.onTileHealthUpdate(x, y, isBackground, health) }
		}

		abstract inner class AbstractTicket(val targetState: ServerChunk.State) : ITicket {
			final override val id: Int = nextTicketID++
			final override val pos: ChunkPos
				get() = this@TicketList.pos

			final override var isCanceled: Boolean = false
			final override val chunk = CompletableFuture<ServerChunk>()

			init {
				isBusy = true
				this@TicketList.targetState.trySend(targetState)
			}

			final override fun cancel() {
				if (isCanceled) return

				ticketListLock.withLock {
					if (isCanceled) return
					isCanceled = true
					chunk.cancel(false)
					listener = null
					cancel0()
				}
			}

			protected abstract fun cancel0()
			final override var listener: IChunkListener? = null
		}

		inner class Ticket(state: ServerChunk.State) : AbstractTicket(state) {
			init {
				permanent.add(this)
				loadChunk()
			}

			override fun cancel0() {
				permanent.remove(this)
			}
		}

		inner class TimedTicket(expiresAt: Int, state: ServerChunk.State) : AbstractTicket(state), ITimedTicket {
			var expiresAt = expiresAt + ticks

			override val timeRemaining: Int
				get() = (expiresAt - ticks).coerceAtLeast(0)

			init {
				temporary.add(this)
				loadChunk()
			}

			override fun cancel0() {
				temporary.remove(this)
			}

			override fun prolong(ticks: Int) {
				if (ticks == 0 || isCanceled) return

				ticketListLock.withLock {
					if (isCanceled) return

					temporary.remove(this)
					expiresAt += ticks
					if (timeRemaining > 0) temporary.add(this)
				}
			}
		}
	}

	@JsonFactory
	data class MetadataJson(
		val playerStart: Vector2d,
		val respawnInWorld: Boolean,
		val adjustPlayerStart: Boolean,
		val worldTemplate: JsonObject,
		val centralStructure: WorldStructure,
		val protectedDungeonIds: IntArraySet,
		val worldProperties: JsonObject,
		val spawningEnabled: Boolean
	)

	companion object {
		private val LOGGER = LogManager.getLogger()

		fun create(server: StarboundServer, template: WorldTemplate, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): ServerWorld {
			return ServerWorld(server, template, storage, worldID)
		}

		fun create(server: StarboundServer, geometry: WorldGeometry, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): ServerWorld {
			return ServerWorld(server, WorldTemplate(geometry), storage, worldID)
		}

		fun create(server: StarboundServer, metadata: WorldStorage.Metadata, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): ServerWorld {
			return AssetPathStack("/") { _ ->
				val meta = Starbound.gson.fromJson(metadata.data.content, MetadataJson::class.java)

				val world = ServerWorld(server, WorldTemplate.fromJson(meta.worldTemplate), storage, worldID)
				world.playerSpawnPosition = meta.playerStart
				world.respawnInWorld = meta.respawnInWorld
				world.adjustPlayerSpawn = meta.adjustPlayerStart
				world.centralStructure = meta.centralStructure

				for ((k, v) in meta.worldProperties.entrySet()) {
					world.setProperty(k, v)
				}

				world.protectedDungeonIDs.addAll(meta.protectedDungeonIds)
				world
			}
		}

		fun load(server: StarboundServer, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): CompletableFuture<ServerWorld> {
			LOGGER.info("Attempting to load world at $worldID")

			return storage.loadMetadata().thenApply {
				LOGGER.info("Loading world at $worldID")
				AssetPathStack("/") { _ ->
					val meta = it.map { Starbound.gson.fromJson(it.data.content, MetadataJson::class.java) }.orThrow { NoSuchElementException("No world metadata is present") }

					val world = ServerWorld(server, WorldTemplate.fromJson(meta.worldTemplate), storage, worldID)
					world.playerSpawnPosition = meta.playerStart
					world.respawnInWorld = meta.respawnInWorld
					world.adjustPlayerSpawn = meta.adjustPlayerStart
					world.centralStructure = meta.centralStructure

					for ((k, v) in meta.worldProperties.entrySet()) {
						world.setProperty(k, v)
					}

					world.protectedDungeonIDs.addAll(meta.protectedDungeonIds)
					world
				}
			}.exceptionally {
				LOGGER.error("Error while instancing world $worldID", it)
				null
			}
		}
	}
}
